from pandas import DataFrame, Series, Index

from norm.executable import NormExecutable
from norm.executable.schema.declaration import TypeDeclaration
from norm.executable.expression import NormExpression
from norm.models import Lambda, Status

import logging
logger = logging.getLogger(__name__)


class TypeImplementation(NormExecutable):

    def __init__(self, type_, op, query, description):
        """
        The implementation of a type. Depending on the operation, it can be initial implementation or
        incremental implementation.
        :param type_: the type to implement
        :type type_: TypeDeclaration
        :param op: the operation of the implementation, i.e., [':=', '|=', '&=']
        :type op: ImplType
        :param query: the query expression that implements the type
        :type query: NormExpression
        :param description: the implementation documentation. it will be appended to original one for incremental.
        :type description: str
        """
        super().__init__()
        self.type_ = type_
        self.op = op
        self.query = query
        self.description = description

    def compile(self, context):
        """
        Three types of implementations
            * new implementation (:=) removes all revisions from this version
            * conjunctive implementation (&=)
            * disjunctive implementation (|=)
        """
        lam = self.type_.lam
        if lam.status != Status.DRAFT:
            logger.info('Lambda: {} is not in DRAFT mode. Import first'.format(lam))
            lam = Lambda(namespace=context.context_namespace, name=lam.name, description=lam.description,
                         variables=lam.variables, user=context.user)
        if self.description is not None and self.description.strip() != '':
            self.query.description = self.description
        self.lam = lam
        return self

    def execute(self, context):
        from norm.engine import ImplType
        from norm.models.norm import RevisionType
        lam = self.lam
        delta = self.query.execute(context)
        if isinstance(delta, DataFrame):
            if self.query.lam is not self.lam and self.lam is not self.query.lam.cloned_from:
                # reset the index name if the output lambda is not the lambda to be revised
                delta.index.name = ''

            if delta.index.name != lam.VAR_OID:
                delta = lam.fill_primary(delta)
                delta = lam.fill_time(delta)
                delta = lam.fill_oid(delta)
        elif isinstance(delta, Index):
            assert(delta.name == lam.VAR_OID)
            cols = [v.name for v in lam.variables if v.name in self.query.lam]
            delta = self.query.lam.data.loc[delta, cols]
            delta = lam.fill_primary(delta)
            delta = lam.fill_time(delta)
            delta = lam.fill_oid(delta.reset_index(drop=True))
        qs = str(self.query)
        if self.op == ImplType.DEF:
            # If the query already exists, the revision is skipped
            if any((rev.query == qs for rev in lam.revisions)):
                return lam

            # Ensure the Lambda is a new implementation
            if len(lam.revisions) > 0:
                lam.remove_revisions()
            lam.revise(qs, self.description, delta, RevisionType.DISJUNCTION)
        elif self.op == ImplType.OR_DEF:
            lam.revise(qs, self.description, delta, RevisionType.DISJUNCTION)
        elif self.op == ImplType.AND_DEF:
            if isinstance(delta, Series):
                if len(delta) == 1:
                    vname = delta.name
                    vvalue = delta.values[0]
                    delta = DataFrame(index=lam.data.index)
                    delta[vname] = vvalue
                    delta = lam.fill_time(delta)
                elif delta.index.name == lam.VAR_OID:
                    delta = DataFrame(delta)
                    delta = lam.fill_time(delta)
            elif isinstance(delta, DataFrame):
                delta = lam.fill_time(delta)
            lam.revise(qs, self.description, delta, RevisionType.CONJUNCTION)
        return lam



import uuid

from norm.grammar.literals import AOP
from norm.executable import NormError
from norm.executable.expression import NormExpression
from norm.executable.expression.evaluation import EvaluationExpr

import logging

logger = logging.getLogger(__name__)


class ArithmeticExpr(NormExpression):

    def __init__(self, op, expr1, expr2=None, projection=None):
        """
        Arithmetic expression
        :param op: the operation, e.g., [+, -, *, /, %, **]
        :type op: AOP
        :param expr1: left expression
        :type expr1: ArithmeticExpr
        :param expr2: right expression
        :type expr2: ArithmeticExpr
        """
        super().__init__()
        self.op: AOP = op
        self.expr1: ArithmeticExpr = expr1
        self.expr2: ArithmeticExpr = expr2
        self.projection = projection
        self._projection_name = None
        assert(self.op is not None)
        assert(self.expr2 is not None)
        self._exprstr: str = None

    def __str__(self):
        if self._exprstr is None:
            msg = 'Compile the expression first'
            logger.error(msg)
            raise NormError(msg)
        return self._exprstr

    def compile(self, context):
        if self.expr1:
            if isinstance(self.expr1, EvaluationExpr) and self.expr1.variable is None and len(self.expr1.args) == 1:
                # It is possible to be parsed as arguments
                self.expr1 = self.expr1.args[0].expr
        if isinstance(self.expr2, EvaluationExpr) and self.expr2.variable is None and len(self.expr2.args) == 1:
            # It is possible to be parsed as arguments
            self.expr2 = self.expr2.args[0].expr

        if self.op is AOP.SUB and self.expr1 is None:
            self._exprstr = '-({})'.format(self.expr2)
        else:
            self._exprstr = '({}) {} ({})'.format(self.expr1, self.op, self.expr2)
        self.eval_lam = context.scope
        from norm.models import Lambda, Variable, lambdas
        if self.projection and self.projection.num > 0:
            self._projection_name = self.projection.variables[0].name
        else:
            self._projection_name = Lambda.VAR_OUTPUT
        self.lam = Lambda(context.context_namespace, context.TMP_VARIABLE_STUB + str(uuid.uuid4()),
                          variables=[Variable(self._projection_name, lambdas.Any)])
        self.lam.cloned_from = self.eval_lam
        return self

    def execute(self, context):
        df = self.eval_lam.data.eval(self._exprstr)
        if self.projection and self.projection.num > 0:
            from pandas import DataFrame, Series
            if isinstance(df, DataFrame):
                raise NormError('cant be here')
                self.data = df.renames({old_name: new_var.name for old_name, new_var in
                                        zip(df.columns, self.projection.variables)})
        assert(isinstance(df, Series))
        self.lam.data = DataFrame({self._projection_name: df})
        return self.lam.data



from norm.executable import NormExecutable
from norm.executable.schema.variable import VariableName
from norm.executable.schema.type import TypeName
from norm.models import Lambda, Status, PythonLambda
from typing import List

import logging
logger = logging.getLogger(__name__)


class ArgumentDeclaration(object):

    def __init__(self, variable_name, variable_type, optional=False):
        """
        The argument declaration
        :param variable_name: the name of the variable
        :type variable_name: VariableName
        :param variable_type: the type of the variable
        :type variable_type: TypeName
        :param optional: whether it is optional or not
        :type optional: bool
        """
        assert(variable_type is not None)
        assert(variable_type.lam is not None)
        from norm.models import Variable
        self.var = Variable(variable_name.name, variable_type.lam, not optional)


class RenameArgument(object):

    def __init__(self, variable_original_name, variable_new_name):
        """
        Rename a variable
        :param variable_original_name: the original name
        :type variable_original_name: str
        :param variable_new_name: the new name
        :type variable_new_name: str
        """
        self.variable_original_name = variable_original_name
        self.variable_new_name = variable_new_name


class TypeDeclaration(NormExecutable):

    def __init__(self, type_name, argument_declarations=None, output_type_name=None):
        """
        The type declaration
        :param type_name: the type name
        :type type_name: TypeName
        :param argument_declarations: the list of argument declarations
        :type argument_declarations: List[ArgumentDeclaration]
        :param output_type_name: the type_name as output, default to boolean
        :type output_type_name: TypeName
        """
        super().__init__()
        self.type_name = type_name
        self.argument_declarations = argument_declarations
        self.output_type_name = output_type_name
        self._description = None

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        self._description = value
        self.lam.description = value

    def compile(self, context):
        """
        Declare a type:
            * Create a type
            * Add new variable to a type # TODO
            * Modify description # TODO
            * Modify variables # TODO
        :return: the lambda
        :rtype: Lambda
        """
        # TODO: optimize to query db in batch for all types or utilize cache
        lam = self.type_name.lam  # type: Lambda
        if self.output_type_name is not None:
            output_arg = ArgumentDeclaration(VariableName(None, Lambda.VAR_OUTPUT), self.output_type_name)
            if self.argument_declarations is None:
                self.argument_declarations = [output_arg]
            else:
                self.argument_declarations.append(output_arg)
        if self.argument_declarations is None:
            variables = []
        else:
            variables = [var_declaration.var for var_declaration in self.argument_declarations]

        if lam is None:
            #  Create a new Lambda
            lam = Lambda(namespace=context.context_namespace, name=self.type_name.name)
            lam.description = self.description
            lam.variables = variables
            context.session.add(lam)
            from norm.config import cache
            cache[(context.context_namespace, self.type_name.name, None, None)] = lam
        else:
            assert(lam.status == Status.DRAFT)
            if len(variables) > 0 and \
                    sorted(lam.variables, key=lambda v: v.name) != sorted(variables, key=lambda v: v.name):
                # Revise the existing schema
                new_variables = {v.name: v.type_ for v in variables}
                current_variables = {v.name: v.type_ for v in lam.variables}
                lam.delete_variable([v.name for v in lam.variables if new_variables.get(v.name, None) != v.type_])
                lam.add_variable([v for v in variables if current_variables.get(v.name, None) != v.type_])
                # TODO: make a doc change revision
                if self.description:
                    lam.description = self.description

        self.lam = lam
        return self


class AdditionalTypeDeclaration(TypeDeclaration):

    def __init__(self, type_name, argument_declarations):
        """
        Adding additional variables to the type or override the existing ones
        :param type_name: the name of type
        :type type_name: TypeName
        :param argument_declarations: the list of variables to add or modify
        :type argument_declarations: List[ArgumentDeclaration]
        """
        super().__init__(type_name, argument_declarations)

    def compile(self, context):
        lam = self.type_name.lam
        variables = [var_declaration.var for var_declaration in self.argument_declarations]
        assert(lam.status == Status.DRAFT)
        assert(sorted(lam.variables, key=lambda v: v.name) != sorted(variables, key=lambda v: v.name))
        # Revise the existing schema
        current_variables = {v.name: v.type_ for v in lam.variables}
        lam.add_variable([v for v in variables if v.name not in current_variables.keys()])
        lam.astype([v for v in variables if v.name in current_variables.keys()
                    and current_variables.get(v.name) != v.type_])
        self.lam = lam
        return self


class RenameTypeDeclaration(TypeDeclaration):

    def __init__(self, type_name, rename_arguments):
        """
        Rename the variables
        :param type_name: the name of the type
        :type type_name: TypeName
        :param rename_arguments: the list of variables to rename
        :type rename_arguments: List[RenameArgument]
        """
        super().__init__(type_name)
        self.rename_arguments = rename_arguments

    def compile(self, context):
        lam = self.type_name.lam
        assert(lam is not None)
        variables = {rename.variable_original_name: rename.variable_new_name for rename in self.rename_arguments
                     if rename.variable_original_name in lam}
        lam.rename_variable(variables)
        self.lam = lam
        return self


class CodeTypeDeclaration(TypeDeclaration):

    def __init__(self, type_name, code, description):
        """
        Declare a python lambda
        :param type_name: the name of the lambda
        :type type_name: TypeName
        :param code: the python code
        :type code: str
        :param description: the description of the lambda
        :type description: str
        """
        super().__init__(type_name)
        self.code = code
        self._description = description

    def compile(self, context):
        lam = PythonLambda(context.context_namespace, self.type_name.name, self.description, self.code)
        context.session.add(lam)
        from norm.config import cache
        cache[(context.context_namespace, self.type_name.name, None, None)] = lam
        self.lam = lam
        return self



from norm.executable.expression import NormExpression


class CodeExpr(NormExpression):

    def __init__(self, code):
        """
        Evaluate a piece of code in python
        :param code: a piece of code in string
        :type code: str
        """
        super().__init__()
        self.code = code

    def serialize(self):
        pass




from sqlalchemy import Column, Integer, String, Text, exists, and_, TypeDecorator

import json
import logging
import traceback
logger = logging.getLogger(__name__)


def lazy_property(f):
    internal_property = '__lazy_' + f.__name__

    @property
    def lazy_property_wrapper(self):
        try:
            return getattr(self, internal_property)
        except AttributeError:
            v = f(self)
            setattr(self, internal_property, v)
            return v
        except Exception as e:
            raise e
    return lazy_property_wrapper


class LazyMixin(object):
    def invalidate(self, property_name=None):
        """
        Invalidate a lazy property or all lazy properties in the object. If the lazy property does not exist,
        raise AttributeError.
        :param property_name: a lazy property
        :type property_name: str
        :raise: AttributeError
        """
        if property_name is None:
            lazy_properties = (name for name in dir(self) if name.startswith('__lazy_'))
            for p in lazy_properties:
                delattr(self, p)
        else:
            lazy_property_name = '__lazy_' + property_name
            if hasattr(self, lazy_property_name):
                delattr(self, lazy_property_name)


class register_parameter(object):
    PARAMETER_REGISTRY = '__parameter_defaults__'

    def __init__(self, section='main', name='', default='', parameter_type=None, description='', target_class=None):
        self.section = section
        self.name = name
        self.default = default
        if parameter_type is None:
            self.parameter_type = type(default)
        else:
            self.parameter_type = parameter_type
        self.description = description
        self.target_class = target_class

    def __call__(self, cls):
        if self.target_class is not None:
            tcls = self.target_class
        else:
            tcls = cls
        section = cls.__name__ + ':' + self.section
        field = '{}{}'.format(self.PARAMETER_REGISTRY, tcls.__name__)

        if not hasattr(tcls, field):
            # TODO:  decide whether should use dir(cls) or dir(tcls)
            parent_fields = [field for field in dir(cls) if field.startswith(register_parameter.PARAMETER_REGISTRY)]
            parent_param_names = set()
            parent_params = []
            for f in parent_fields:
                p_params=getattr(cls, f)
                # use append then sum should be faster if the loop is deep.  Otherwise, extend should be fast.
                parent_params.append([p for p in p_params if p['name'] not in parent_param_names])
                parent_param_names |= {p['name'] for p in p_params}
            parent_params=sum(parent_params, [])

            setattr(tcls, field, parent_params)
        params = getattr(tcls, field)
        found = [p for p in params if p['name'] == self.name]
        if found:
            msg = 'Parameter {} already defined in class {}'.format(self.name, found[0]['cls'])
            logger.error(msg)
            logger.debug(traceback.print_exc())
            raise ValueError(msg)
        params.append({'section': section,
                       'name': self.name,
                       'default': self.default,
                       'type': self.parameter_type,
                       'description': self.description,
                       'cls': tcls.__name__})
        return cls


class ParametrizedMixin(LazyMixin):
    params = Column(Text, default='{}')

    @lazy_property
    def parameters(self):
        if not self.params or self.params == '':
            return {}

        try:
            return json.loads(self.params)
        except:
            msg = 'Parameters {} can not be parsed as json'.format(self.params)
            logger.error(msg)
            logger.debug(traceback.print_exc())
            raise ValueError(msg)

    @classmethod
    def registered_parameters(cls):
        field = '{}{}'.format(register_parameter.PARAMETER_REGISTRY, cls.__name__)
        if not hasattr(cls, field):
            parent_fields = [field for field in dir(cls) if field.startswith(register_parameter.PARAMETER_REGISTRY)]
            parent_param_names = set()
            parent_params = sum([[p for p in getattr(cls, f) if p['name'] not in parent_param_names]
                                for f in parent_fields], [])
            setattr(cls, field, parent_params)
        return getattr(cls, field)

    @classmethod
    def parameter_default(cls, name):
        """
        Get the default value for a parameter
        :param name: the name of the parameter
        :type name: str
        :return: the default value registered
        """
        for p in cls.registered_parameters():
            if p['name'] == name:
                return p['default']
        return None

    @classmethod
    def parameter_type(cls, name):
        """
        Get the type for a parameter
        :param name: the name of the parameter
        :type name: str
        :return: the type registered
        """
        for p in cls.registered_parameters():
            if p['name'] == name:
                return p['type']
        return None

    @classmethod
    def df_registered_parameters(cls, section=None):
        from pandas import DataFrame
        df = DataFrame(cls.registered_parameters())
        if section is None:
            return df
        else:
            return df.query('section=="{}"'.format(section))

    @classmethod
    def json_registered_parameters(cls, section=None):
        from pandas import DataFrame
        df = DataFrame(cls.registered_parameters())
        if df.empty:
            return '[]'

        df = df[['section', 'name', 'description', 'default']]
        if section is None:
            return df.to_json(orient='records')
        else:
            return df.query('section=="{}"'.format(section)).to_json(orient='records')

    @classmethod
    def markup_registered_parameters(cls, section=None):
        rp = cls.registered_parameters()
        if rp is None or len(rp) == 0:
            return "[]"

        html = "<table width=90%><tr><th>Section</th><th>Name</th><th>Default</th><th>Description</th></tr>"
        for row in rp:
            if section is None or row['section'] == section:
                html += "<tr><td>{}&nbsp;&nbsp;&nbsp;</td><td>{}&nbsp;&nbsp;&nbsp;</td><td>{}&nbsp;&nbsp;&nbsp;</td>" \
                        "<td>{}&nbsp;&nbsp;&nbsp;</td></tr>"\
                    .format(row['section'], row['name'], row['default'], row['description'])
        html += "</table>"
        return html

    def get(self, name, default=None):
        """
        Get the value for a parameter by name
        :param name: the name of the parameter
        :type name: str
        :param default: a given overridden default value
        :return: the value
        """
        return self.parameters.get(name, default or self.parameter_default(name))

    def set(self, name, value):
        """
        Set the value for a parameter
        :param name: the name of the parameter
        :type name: str
        :param value: the value to set
        :return: self
        """
        registered_type = self.parameter_type(name)
        if registered_type is None:
            msg = 'Parameter {} is not registered for class {}'.format(name, self.__class__.__name__)
            logger.error(msg)
            logger.debug(traceback.print_exc())
            raise ValueError(msg)
        if not isinstance(value, registered_type) and not isinstance(registered_type, type(None)):
            msg = 'The value {} is not the same type as registered {}'.format(value, registered_type)
            logger.error(msg)
            logger.debug(traceback.print_exc())
            raise ValueError(msg)
        self.parameters[name] = value
        self.invalidate(name)
        self.params = json.dumps(self.parameters)
        return self


class ARRAY(TypeDecorator):
    """ Sqlite-like does not support arrays.
        Let's use a custom type decorator.

        See http://docs.sqlalchemy.org/en/latest/core/types.html#sqlalchemy.types.TypeDecorator
    """

    impl = String

    def __init__(self, intern=None):
        super().__init__()
        self.intern = intern

    def process_bind_param(self, value, dialect):
        return json.dumps(value)

    def process_result_value(self, value, dialect):
        return json.loads(value)

    def copy(self):
        return ARRAY(self.impl.length)




"""A collection of ORM sqlalchemy models for Lambda"""
import zlib

import norm.config as config
from norm.models.mixins import ParametrizedMixin, ARRAY
from norm.models.license import License
from norm.models.user import User
from norm.models import Model

from sqlalchemy import Column, Integer, String, ForeignKey, Text, Boolean, DateTime, Enum, desc, UniqueConstraint, orm
from sqlalchemy import Table
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship
from sqlalchemy.ext.orderinglist import ordering_list

import json
import os
import errno
import enum

from datetime import datetime
from pandas import DataFrame
from hashids import Hashids
import numpy as np

from typing import List, Dict

import logging
logger = logging.getLogger(__name__)

metadata = Model.metadata
hashids = Hashids(min_length=config.VERSION_MIN_LENGTH)


class Variable(Model):
    """Variable placeholder"""

    __tablename__ = 'variables'

    id = Column(Integer, primary_key=True)
    name = Column(String(256), default='')
    primary = Column(Boolean, default=False)
    position = Column(Integer)
    type_id = Column(Integer, ForeignKey('lambdas.id'))
    type_ = relationship('Lambda', foreign_keys=[type_id])

    def __init__(self, name, type_, primary=False):
        """
        Construct the variable
        :param name: the full name of the variable
        :type name: str
        :param type_: the type of the variable
        :type type_: Lambda
        :param primary: whether the variable is a primary for an Lambda
        :type primary: bool
        """
        self.id = None
        self.name = name
        self.type_ = type_
        self.primary = primary

    def __hash__(self):
        return hash(self.name) + hash(self.type_)

    def __eq__(self, other):
        if other is None or not isinstance(other, Variable):
            return False
        return other.name == self.name and other.type_ == self.type_

    def __repr__(self):
        return '{}: {}'.format(self.name, self.type_)

    def __str__(self):
        return self.__repr__()


lambda_variable = Table(
    'lambda_variable', metadata,
    Column('id', Integer, primary_key=True),
    Column('lambda_id', Integer, ForeignKey('lambdas.id')),
    Column('variable_id', Integer, ForeignKey('variables.id')),
)


class Status(enum.Enum):
    """
    Indicate whether the lambda function can be modified or not
    """
    DRAFT = 0
    READY = 1


class RevisionType(enum.Enum):
    """
    Revision type
    """
    DISJUNCTION = 0
    CONJUNCTION = 1


def new_version():
    import time
    return '$' + hashids.encode(int(time.time() * 1000))


def _check_draft_status(func):
    """
    A decorator to check whether the current Lambda is in draft status
    :param func: a function to wrap
    :type func: Callable
    :return: a wrapped function
    :rtype: Callable
    """
    def wrapper(self, *args, **kwargs):
        if self.status != Status.DRAFT:
            msg = '{} is not in Draft status. Please clone first to modify'.format(self)
            logger.error(msg)
            raise RuntimeError(msg)
        return func(self, *args, **kwargs)
    return wrapper


def _only_queryable(func):
    """
    A decorator to bypass the function if the current Lambda is below queryable
    :param func: a function to wrap
    :type func: Callable
    :return: a wrapped function
    :rtype: Callable
    """
    def wrapper(self, *args, **kwargs):
        if not self.queryable:
            msg = '{} is not Queryable. Need to add data.'.format(self)
            logger.error(msg)
            raise RuntimeError(msg)
        return func(self, *args, **kwargs)
    return wrapper


def _only_adaptable(func):
    """
    A decorator to bypass the function if the current Lambda is below adaptable
    :param func: a function to wrap
    :type func: Callable
    :return: a wrapped function
    :rtype: Callable
    """
    def wrapper(self, *args, **kwargs):
        if not self.adaptable:
            msg = '{} is not adaptable. Need to add probabilistic lambdas.'.format(self)
            logger.error(msg)
            raise RuntimeError(msg)
        return func(self, *args, **kwargs)
    return wrapper


class Lambda(Model, ParametrizedMixin):
    """Lambda model is a function"""
    __tablename__ = 'lambdas'
    category = Column(String(128))

    VAR_OUTPUT = 'output'
    # OUTPUT default to the OID, hence the type is integer
    VAR_LABEL = 'label'
    VAR_LABEL_T = 'float'
    VAR_OID = 'oid'
    VAR_OID_T = 'int64'
    VAR_PROB = 'prob'
    VAR_PROB_T = 'float'
    VAR_TIMESTAMP = 'timestamp'
    VAR_TIMESTAMP_T = 'datetime64[ns]'
    VAR_TOMBSTONE = 'tombstone'
    VAR_TOMBSTONE_T = 'bool'
    VAR_ANONYMOUS_STUB = 'var'

    # identifiers
    id = Column(Integer, primary_key=True)
    namespace = Column(String(512), default='')
    name = Column(String(256), nullable=False)
    # data type
    dtype = Column(String(16), default='object')
    # computational properties
    atomic = Column(Boolean, default=False)
    queryable = Column(Boolean, default=False)
    adaptable = Column(Boolean, default=False)
    # owner
    created_by_id = Column(Integer, ForeignKey(User.id))
    owner = relationship(User, backref='lambdas', foreign_keys=[created_by_id])
    # audition
    created_on = Column(DateTime, default=datetime.now)
    changed_on = Column(DateTime, default=datetime.now, onupdate=datetime.now)
    # schema
    description = Column(Text, default='')
    variables = relationship(Variable, order_by=Variable.position, collection_class=ordering_list('position'),
                             secondary=lambda_variable)
    # version
    anchor = Column(Boolean, default=True)
    cloned_from_id = Column(Integer, ForeignKey('lambdas.id'))
    cloned_from = relationship('Lambda', remote_side=[id])
    merged_from_ids = Column(ARRAY(Integer))
    version = Column(String, default=new_version, nullable=False)
    # revision
    revisions = relationship('Revision', order_by='Revision.position', collection_class=ordering_list('position'))
    current_revision = Column(Integer, default=-1)
    status = Column(Enum(Status), default=Status.DRAFT)
    # license
    license_id = Column(Integer, ForeignKey(License.id))
    license = relationship(License, foreign_keys=[license_id])

    __mapper_args__ = {
        'polymorphic_identity': 'lambda',
        'polymorphic_on': category,
        'with_polymorphic': '*'
    }

    __table_args__ = tuple(UniqueConstraint('namespace', 'name', 'version', name='unique_lambda'))

    def __init__(self, namespace='', name='', description='', params='{}', variables=None, dtype=None, user=None):
        self.id: int = None
        self.namespace: str = namespace
        self.name: str = name
        self.version: str = new_version()
        self.description: str = description
        self.params: str = params
        self.owner: User = user
        self.status: Status = Status.DRAFT
        self.merged_from_ids: List[int] = []
        self.variables: List[Variable] = variables or []
        from norm.models.revision import Revision
        self.revisions: List[Revision] = []
        self.current_revision: int = -1
        self.dtype: str = dtype or 'object'
        self.anchor: bool = True
        self.atomic: bool = False
        self.queryable: bool = False
        self.adaptable: bool = False
        self._data: DataFrame = None
        self.modified_or_new = True
        self.license_id = 0

    @orm.reconstructor
    def init_on_load(self):
        self._data = None
        self.modified_or_new = False

    @hybrid_property
    def nargs(self):
        return len(self.variables)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return self.signature

    def __hash__(self):
        if self.id is not None:
            return self.id
        h = 0
        if self.namespace:
            h += hash(self.namespace) << 7
        h += hash(self.name) << 7
        if self.version:
            h += hash(self.version)
        return h

    def __eq__(self, other):
        if not isinstance(other, Lambda):
            return False

        return hash(self) == hash(other)

    def __contains__(self, item):
        if isinstance(item, Variable):
            return item.name in self._all_columns

        if isinstance(item, str):
            return item in self._all_columns

        return False

    def __call__(self, **kwargs):
        pass

    def _repr_html_(self):
        return self.data._repr_html_()

    @classmethod
    def exists(cls, session, obj):
        """
        Build the SQLAlchemy equality condition
        :param session: the session to check against
        :type session: sqlalchemy.orm.Session
        :param obj: another Lambda to compare with
        :type obj: Lambda
        :return: List
        """
        in_stores = session.query(Lambda)\
                           .filter(Lambda.name == obj.name,
                                   Lambda.description == obj.description,
                                   Lambda.namespace == obj.namespace)\
                           .all()
        return any(in_store.variables == obj.variables for in_store in in_stores)

    @property
    def default(self):
        """
        The default value for the Lambda
        """
        return {v.name: v.type_.default for v in self.variables}

    @property
    def is_functional(self):
        """
        Whether the Lambda is in the relational or functional format
        :rtype: bool
        """
        return len(self.variables) > 0 and self.variables[-1].name == self.VAR_OUTPUT

    @property
    def output_type(self):
        """
        Return the output type. If no output variable defined, return itself
        :return: the output type
        :rtype: Lambda
        """
        if self.is_functional:
            return self.variables[-1].type_
        else:
            return self

    def get_type(self, variable_name):
        """
        Get the type of the variable by the name
        :param variable_name: the name of the variable
        :type variable_name: str
        :return: the type of that variable
        :rtype: Lambda
        """
        if variable_name == self.VAR_OUTPUT and not self.is_functional:
            return self
        elif variable_name == self.VAR_OID:
            return self.VAR_OID_T
        elif variable_name == self.VAR_TIMESTAMP:
            return self.VAR_TIMESTAMP_T
        elif variable_name == self.VAR_PROB:
            return self.VAR_PROB_T
        elif variable_name == self.VAR_LABEL:
            return self.VAR_LABEL_T
        elif variable_name == self.VAR_TOMBSTONE:
            return self.VAR_TOMBSTONE_T

        for v in self.variables:
            if v.name == variable_name:
                return v.type_
        return None

    @property
    def data(self):
        if self._data is None:
            if self.queryable:
                self._data = self._load_data()
            else:
                self._data = self.empty_data()
        return self._data

    @data.setter
    def data(self, value):
        if value is None:
            self._data = self.empty_data()
        else:
            self._data = value

    @property
    def signature(self):
        if self.namespace is not None and self.namespace.strip() != '':
            return self.namespace + '.' + self.name + self.version[:config.VERSION_MIN_LENGTH]
        else:
            return self.name + self.version[:config.VERSION_MIN_LENGTH]

    def clone(self):
        """
        Clone itself and bump up the version. Make sure updates are done after clone.
        :return: the cloned version of it
        :rtype: Lambda
        """
        lam = self.__class__(namespace=self.namespace, name=self.name, description=self.description,
                             params=self.params, variables=self.variables)
        lam.cloned_from = self
        lam.anchor = False
        lam.atomic = self.atomic
        lam.queryable = self.queryable
        lam.adaptable = self.adaptable
        if self._data is not None:
            lam._data = self._data
        return lam

    def merge(self, others):
        """
        Clone itself and merge several other versions into the new version
        :param others: other versions
        :type others: List[Lambda]
        :return: the merged version
        :rtype: Lambda
        """
        assert(all([o.namespace == self.namespace and o.name == self.name for o in others]))

        lam = self.clone()
        lam.merged_from_ids = [o.id for o in others]
        for o in others:
            lam.revisions.extend(o.revisions)
        while not self.end_of_revisions:
            self.forward()
        lam.status = Status.READY
        return lam

    def compact(self):
        """
        Compact this version with previous versions to make an anchor version
        :return:
        TODO: to implement
        """
        pass

    def _add_optional_variables(self, df):
        from norm.models.native import get_type_by_dtype
        current_variable_names = set(self._all_columns)
        vars_to_add = [Variable(col, get_type_by_dtype(dtype)) for col, dtype in zip(df.columns, df.dtypes)
                       if col not in current_variable_names]
        if len(vars_to_add) > 0:
            from norm.models.revision import AddVariableRevision
            self._add_revision(AddVariableRevision(vars_to_add, self))

    @_check_draft_status
    def revise(self, query, description, df, op):
        """
        Revise the Lambda by a given query and its result. If the query is an action, result is None.
        :param query: the given query string grounded with specific versions
        :type query: str
        :param description: the comment on the revision
        :type description: str
        :param df: the result data
        :type df: DataFrame
        :param op: the type of revision, conjunction or disjunction
        :type op: RevisionType
        :return: revised data or None
        :rtype: DataFrame
        """
        # If the query already exists, the revision is skipped
        if any((rev.query == query for rev in self.revisions)):
            return self.data.loc[df.index] if df is not None else None

        if df is not None and isinstance(df, DataFrame):
            assert(df.index.name == self.VAR_OID)
            assert(self.VAR_TIMESTAMP in df.columns)

            # Add new optional variables if the data contains variables that are not declared in this Lambda
            self._add_optional_variables(df)

            # Drop duplicated rows by the index
            delta = df[~df.index.duplicated(keep='first')]
        else:
            delta = None

        # Apply corresponding revisions
        from norm.models.revision import ConjunctionRevision, DisjunctionRevision
        if op == RevisionType.DISJUNCTION:
            self._add_revision(DisjunctionRevision(query, description, self, delta))
        elif op == RevisionType.CONJUNCTION:
            self._add_revision(ConjunctionRevision(query, description, self, delta))
        else:
            msg = 'Revision {} has not been implemented yet'.format(op)
            logger.error(msg)
            raise NotImplementedError(msg)

        # Enable queryable
        if not self.queryable:
            self.queryable = True

        return self.data.loc[df.index] if delta is not None else None

    @_check_draft_status
    @_only_adaptable
    def fit(self):
        """
        Fit the model with all the existing data or given data
        """
        from norm.models.revision import FitRevision
        # TODO: implement the query
        revision = FitRevision('', '', self)
        self._add_revision(revision)

    def fill_na(self, df):
        """
        Fill the null data with default values
        :param df: the data
        :type df: DataFrame
        :return: the filled DataFrame
        :rtype: DataFrame
        """
        df = self._fill_label(df)
        df = self._fill_prob(df)
        df = self._fill_tombstone(df)
        return df

    def fill_primary(self, df):
        for var in self.variables:
            if var.primary:
                if var.name in df.columns:
                    df.loc[:, var.name] = df[var.name].fillna(var.type_.default)
                else:
                    df[var.name] = var.type_.default
        return df

    def fill_oid(self, df):
        if self.VAR_OID in df.columns or df.index.name == self.VAR_OID:
            return df

        cols = [v.name for v in self.variables if v.primary and v.name in df.columns]
        c = None
        for col in cols:
            if c is None:
                c = df[col].astype(str)
            else:
                c = c.str.cat(df[col].astype(str))
        if c is not None:
            df[self.VAR_OID] = c.astype('bytes').apply(zlib.adler32).astype('int64')
            df = df.set_index(self.VAR_OID)
        else:
            df.index.rename(self.VAR_OID, inplace=True)
        return df

    def fill_time(self, df):
        if self.VAR_TIMESTAMP not in df.columns:
            df[self.VAR_TIMESTAMP] = np.datetime64(datetime.utcnow())
        else:
            df[self.VAR_TIMESTAMP].fillna(np.datetime64(datetime.utcnow()), inplace=True)
        return df

    def _fill_label(self, df):
        if self.VAR_LABEL not in df.columns:
            df[self.VAR_LABEL] = 1.0
        else:
            df[self.VAR_LABEL].fillna(1.0, inplace=True)
        return df

    def _fill_prob(self, df):
        if self.VAR_PROB not in df.columns:
            df[self.VAR_PROB] = 1.0
        else:
            df[self.VAR_PROB].fillna(1.0, inplace=True)
        return df

    def _fill_tombstone(self, df):
        if self.VAR_TOMBSTONE not in df.columns:
            df[self.VAR_TOMBSTONE] = False
        else:
            df[self.VAR_TOMBSTONE].fillna(False, inplace=True)
        return df

    @_check_draft_status
    def add_variable(self, variables):
        """
        Add new new variables to the Lambda
        :type variables: List[Variable] or Variable
        """
        if isinstance(variables, Variable):
            variables = [variables]
        if len(variables) == 0:
            return
        from norm.models.revision import AddVariableRevision
        revision = AddVariableRevision(variables, self)
        self._add_revision(revision)

    @_check_draft_status
    def delete_variable(self, names):
        """
        Delete variables from the Lambda
        :type names: List[str]
        """
        if len(names) == 0:
            return
        from norm.models.revision import DeleteVariableRevision
        revision = DeleteVariableRevision(names, self)
        self._add_revision(revision)

    @_check_draft_status
    def rename_variable(self, renames):
        """
        Change a variable name to another. The argument is a on keyword argument format. The key should exist in
        the Lambda and the value to be the target name.
        :type renames: Dict[str, str]
        """
        if len(renames) == 0:
            return
        from norm.models.revision import RenameVariableRevision
        revision = RenameVariableRevision(renames, self)
        self._add_revision(revision)

    @_check_draft_status
    def astype(self, variables):
        """
        Change the type of variables. The variable names to be changed should exist in current Lambda. The new types
        are specified in the variable type_ attribute.
        :type variables: List[Variable]
        """
        if len(variables) == 0:
            return
        from norm.models.revision import RetypeVariableRevision
        revision = RetypeVariableRevision(variables, self)
        self._add_revision(revision)

    def _add_revision(self, revision):
        revision.apply()
        self.current_revision += 1
        self.modified_or_new = True

    @_check_draft_status
    def save(self):
        """
        Save the current version
        """
        if self.queryable:
            self._save_data()
        if self.adaptable:
            self._save_model()
        self.modified_or_new = False

    @_check_draft_status
    def remove_revisions(self):
        self.current_revision = -1
        self.revisions = []

    @property
    def empty_revisions(self):
        return len(self.revisions) == 0

    @property
    def end_of_revisions(self):
        return self.current_revision == len(self.revisions) - 1

    @property
    def begin_of_revisions(self):
        return self.current_revision == -1

    @_check_draft_status
    def rollback(self):
        """
        Rollback to the previous revision if it is in draft status
        """
        if 0 <= self.current_revision < len(self.revisions):
            self.revisions[self.current_revision].undo()
            self.current_revision -= 1
        else:
            if self.current_revision == -1:
                msg = '{} has already rolled back to the beginning of the version.\n ' \
                      'Might want to rollback the version'.format(self)
                logger.warning(msg)
            else:
                msg = 'Current revision {} is higher than it has {}'.format(self.current_revision, len(self.revisions))
                logger.error(msg)
                raise RuntimeError(msg)

    @_check_draft_status
    def forward(self):
        """
        Forward to the next revision if it is in draft status
        """
        if self.current_revision < len(self.revisions) - 1:
            self.revisions[self.current_revision + 1].redo()
            self.current_revision += 1
        else:
            if self.current_revision == len(self.revisions) - 1:
                msg = '{} is already at the latest revision.\n ' \
                      'Might want to forward the version'.format(self)
                logger.warning(msg)
            else:
                msg = 'Current revision {} is higher than it has {}'.format(self.current_revision, len(self.revisions))
                logger.error(msg)
                raise RuntimeError(msg)

    @property
    def folder(self):
        return '{}/{}/{}/{}'.format(config.DATA_STORAGE_ROOT,
                                    self.namespace.replace('.', '/'),
                                    self.name,
                                    self.version)

    @_only_queryable
    def _create_folder(self):
        """
        Create the folder for the namespace.
        """
        # TODO: abstract the data storage folder creation
        try:
            # for the case of concurrent processing
            os.makedirs(self.folder)
        except OSError as e:
            if e.errno != errno.EEXIST:
                msg = 'Can not create the folder {}'.format(self.folder)
                logger.error(msg)
                logger.error(str(e))
                raise e

    @property
    def _all_columns(self):
        return [self.VAR_OID, self.VAR_PROB, self.VAR_LABEL, self.VAR_TIMESTAMP,
                self.VAR_TOMBSTONE] + [v.name for v in self.variables]

    @property
    def _all_column_types(self):
        return [self.VAR_OID_T, self.VAR_PROB_T, self.VAR_LABEL_T, self.VAR_TIMESTAMP_T,
                self.VAR_TOMBSTONE_T] + [v.type_.dtype for v in self.variables]

    def empty_data(self):
        """
        Create an empty data frame
        :return: the data frame with columns
        :rtype: DataFrame
        """
        df = DataFrame(columns=self._all_columns)
        return df.astype(dict(zip(self._all_columns, self._all_column_types))).set_index(self.VAR_OID)

    @_only_queryable
    def _load_data(self):
        """
        Load data if it exists. If the current version is not an anchor, the previous versions will be combined.
        :return: the combined data
        :rtype: DataFrame
        """
        if self._data is not None:
            return self._data

        if self.anchor:
            self._data = self.empty_data()
        elif self.cloned_from is None:
            msg = "Failed to find the anchor version. The chain is broken for {}".format(self)
            logger.error(msg)
            raise RuntimeError(msg)
        else:
            self._data = self.cloned_from._load_data()

        from norm.models.revision import DeltaRevision
        for i in range(self.current_revision + 1):
            revision = self.revisions[i]
            if isinstance(revision, DeltaRevision):
                revision.redo()

        # Choose the rows still alive and the columns specified in schema
        self._data = self._data[self._all_columns[1:]][~self._data[self.VAR_TOMBSTONE]]
        return self._data

    @_only_queryable
    def _save_data(self):
        """
        Save all revisions' data
        """
        if not os.path.exists(self.folder):
            self._create_folder()

        for revision in self.revisions:
            revision.save()

    @_only_adaptable
    def _build_model(self):
        """
        Build an adaptable model
        TODO: to implement
        """
        pass

    @_only_adaptable
    def _load_model(self):
        """
        Load an adapted model
        TODO: to implement
        :return:
        """
        pass

    @_only_adaptable
    def _save_model(self):
        """
        Save an adapted model
        TODO: to implement
        :return:
        """
        pass


class GroupLambda(Lambda):

    __mapper_args__ = {
        'polymorphic_identity': 'lambda_group'
    }

    def all(self):
        """
        Combine the groups as columns
        :return: Lambda
        """
        raise NotImplementedError

    def any(self):
        """
        Combine the groups as concatenation
        :return: Lambda
        """
        raise NotImplementedError


def retrieve_type(namespaces, name, version=None, status=None):
    """
    Retrieving a Lambda
    :type namespaces: str, List[str] or None
    :type name: str
    :type version: str or None
    :type status: Status or None
    :return: the Lambda or None
    """
    from norm.config import session

    if version == '$lastest':
        version = None
    elif version == '$best':
        raise NotImplementedError('Select the version for the best is not implemented yet')

    queries = [Lambda.name == name]
    if status is not None and isinstance(status, Status):
        queries.append(Lambda.status == status)
    if namespaces is not None:
        if isinstance(namespaces, str):
            queries.append(Lambda.namespace == namespaces)
        else:
            queries.append(Lambda.namespace.in_(namespaces))
    if version is not None:
        queries.append(Lambda.version.startswith(version))

    lam = session.query(Lambda) \
                 .filter(*queries) \
                 .order_by(desc(Lambda.created_on)) \
                 .first()
    return lam






from norm.executable import NormExecutable, NormError
from norm.executable.schema.type import TypeName

from norm.models import Status

import logging

from norm.models.norm import new_version

logger = logging.getLogger(__name__)


class Import(NormExecutable):

    def __init__(self, namespace=None, type_=None, variable=None):
        """
        Import the namespace, if the variable name is given, the imported type is cloned to the
        current context with the variable name
        :param namespace: the namespace
        :type namespace: str
        :param type_: the type
        :type type_: TypeName
        :param variable: the variable
        :type variable: str
        """
        super().__init__()
        assert(namespace is not None)
        assert(namespace != '')
        self.namespace = namespace
        self.type_ = type_
        self.variable = variable

    def compile(self, context):
        """
        Imports follow the following logic:
            * imported namespace is stored in the context
            * imported type is cloned in the context namespace as a draft
            * imported type with alias is cloned and renamed in the context namespace as a draft
        """
        if self.namespace not in context.search_namespaces:
            context.search_namespaces.append(self.namespace)
        if self.type_:
            if self.type_.namespace != self.namespace:
                self.type_.namespace = self.namespace
                self.type_.compile(context)  # require a recompilation since namespace has been changed
            lam = self.type_.lam
            if lam is None:
                msg = "Can not find the type {} in namespace {}".format(self.type_.name, self.namespace)
                logger.error(msg)
                raise NormError(msg)
            if self.variable:
                alias = lam.clone()
                alias.namespace = context.context_namespace
                alias.name = self.variable
                context.session.add(alias)
                self.lam = alias
            else:
                self.lam = lam
        else:
            from norm.models.norm import Lambda
            self.lam = Lambda(self.namespace, '*')
        return self


class Export(NormExecutable):

    def __init__(self, namespace=None, type_=None, alias=None):
        """
        Export the type to the namespace
        :param namespace: the namespace
        :type namespace: str or None
        :param type_: the type
        :type type_: TypeName
        :param alias: the alias in the namespace
        :type alias: str or None
        """
        super().__init__()
        assert(type_ is not None)
        self.namespace = namespace
        self.type_ = type_
        self.alias = alias

    def compile(self, context):
        session = context.session
        lam = self.type_.lam
        if lam is None:
            msg = "Can not find the type {} in namespace {}".format(self.type_.name, self.type_.namespace)
            logger.error(msg)
            raise NormError(msg)
        if lam.status != Status.DRAFT:
            msg = 'Type {} is not in the draft status'
            logger.error(msg)
            raise NormError(msg)

        if self.namespace is None or self.namespace.strip() == '':
            if lam.cloned_from:
                lam.namespace = lam.cloned_from.namespace
            else:
                lam.namespace = context.user_namespace
        else:
            lam.namespace = self.namespace

        old_lam_name = lam.name
        if self.alias:
            lam.name = self.alias
        lam.version = new_version()
        lam.status = Status.READY

        # clone this one back to the current context for further modification
        new_lam = lam.clone()
        new_lam.namespace = context.context_namespace
        new_lam.name = old_lam_name
        session.add(new_lam)
        from norm.config import cache
        cache[(context.context_namespace, new_lam.name, None, None)] = new_lam
        # TODO: possible cascaded exporting. the clone_from object might need to be exported too, or merge into one.
        self.lam = lam
        return self





"""init schema

Revision ID: 467af4a6e6b3
Revises: 
Create Date: 2019-03-11 16:23:21.021301

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '467af4a6e6b3'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    from norm.models.mixins import ARRAY
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('licenses',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=256), nullable=False),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('users',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('first_name', sa.String(length=64), nullable=False),
    sa.Column('last_name', sa.String(length=64), nullable=False),
    sa.Column('username', sa.String(length=64), nullable=False),
    sa.Column('email', sa.String(length=64), nullable=False),
    sa.Column('last_login', sa.DateTime(), nullable=True),
    sa.Column('login_count', sa.Integer(), nullable=True),
    sa.Column('fail_login_count', sa.Integer(), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('email'),
    sa.UniqueConstraint('username')
    )
    op.create_table('lambdas',
    sa.Column('params', sa.Text(), nullable=True),
    sa.Column('category', sa.String(length=128), nullable=True),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('namespace', sa.String(length=512), nullable=True),
    sa.Column('name', sa.String(length=256), nullable=False),
    sa.Column('dtype', sa.String(length=16), nullable=True),
    sa.Column('adaptable', sa.Boolean(), nullable=True),
    sa.Column('atomic', sa.Boolean(), nullable=True),
    sa.Column('queryable', sa.Boolean(), nullable=True),
    sa.Column('created_by_id', sa.Integer(), nullable=True),
    sa.Column('created_on', sa.DateTime(), nullable=True),
    sa.Column('changed_on', sa.DateTime(), nullable=True),
    sa.Column('description', sa.Text(), nullable=True),
    sa.Column('anchor', sa.Boolean(), nullable=True),
    sa.Column('cloned_from_id', sa.Integer(), nullable=True),
    sa.Column('merged_from_ids', ARRAY(), nullable=True),
    sa.Column('version', sa.String(), nullable=False),
    sa.Column('current_revision', sa.Integer(), nullable=True),
    sa.Column('status', sa.Enum('DRAFT', 'READY', name='status'), nullable=True),
    sa.Column('license_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['cloned_from_id'], ['lambdas.id'], ),
    sa.ForeignKeyConstraint(['created_by_id'], ['users.id'], ),
    sa.ForeignKeyConstraint(['license_id'], ['licenses.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('revisions',
    sa.Column('params', sa.Text(), nullable=True),
    sa.Column('category', sa.String(length=128), nullable=True),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('query', sa.Text(), nullable=True),
    sa.Column('position', sa.Integer(), nullable=True),
    sa.Column('description', sa.Text(), nullable=True),
    sa.Column('lambda_id', sa.Integer(), nullable=True),
    sa.Column('renames', sa.JSON(), nullable=True),
    sa.ForeignKeyConstraint(['lambda_id'], ['lambdas.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('variables',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=256), nullable=True),
    sa.Column('primary', sa.Boolean(), nullable=False),
    sa.Column('position', sa.Integer(), nullable=True),
    sa.Column('type_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['type_id'], ['lambdas.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('lambda_variable',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('lambda_id', sa.Integer(), nullable=True),
    sa.Column('variable_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['lambda_id'], ['lambdas.id'], ),
    sa.ForeignKeyConstraint(['variable_id'], ['variables.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('revision_variable',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('revision_id', sa.Integer(), nullable=True),
    sa.Column('variable_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['revision_id'], ['revisions.id'], ),
    sa.ForeignKeyConstraint(['variable_id'], ['variables.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('revision_variable')
    op.drop_table('lambda_variable')
    op.drop_table('variables')
    op.drop_table('revisions')
    op.drop_table('lambdas')
    op.drop_table('users')
    op.drop_table('licenses')
    # ### end Alembic commands ###



from norm.grammar.literals import COP
from norm.executable import Projection
from norm.executable.constant import Constant
from norm.executable.expression import NormExpression

from typing import Union
import logging
logger = logging.getLogger(__name__)


class ArgumentExpr(NormExpression):

    def __init__(self, variable=None, op=None, expr=None, projection=None):
        """
        The argument expression project to a new variable, either assigning or conditional.
        :param variable: the variable
        :type variable: VariableName
        :param expr: the arithmetic expression for the variable
        :type expr: Union[norm.executable.expression.arithmetic.ArithmeticExpr,
                          norm.executable.expression.slice.SliceExpr,
                          norm.executable.expression.evaluation.EvaluationExpr,
                          norm.executable.variable.VariableName]
        :param op: the conditional operation
        :type op: COP
        :param projection: the projection
        :type projection: Projection
        """
        super().__init__()
        from norm.executable.expression.arithmetic import ArithmeticExpr
        from norm.executable.expression.slice import SliceExpr
        from norm.executable.expression.evaluation import EvaluationExpr
        from norm.executable.schema.variable import VariableName
        self.variable: VariableName = variable
        self.expr: Union[ArithmeticExpr, SliceExpr, EvaluationExpr, VariableName] = expr
        self.op: COP = op
        self.projection: Projection = projection

    def __str__(self):
        return '{} {} {} {}'.format(self.variable if self.variable is not None else '',
                                    self.op if self.op is not None else '',
                                    self.expr if self.expr is not None else '',
                                    self.projection if self.projection is not None else '')

    @property
    def is_constant(self):
        return isinstance(self.expr, Constant)

    @property
    def is_assign_operator(self):
        return self.op is None

    def compile(self, context):
        from norm.executable.expression.evaluation import EvaluationExpr
        from norm.executable.schema.variable import VariableName
        if self.is_assign_operator and self.projection is not None and isinstance(self.expr, VariableName):
            self.expr = EvaluationExpr([], self.expr, self.projection).compile(context)
        return self




from norm.config import PUBLIC_USER, session


def login(usr=None):
    from norm.models.user import User
    if usr is None:
        usr = PUBLIC_USER

    u = session.query(User).filter(User.username == usr['username'],
                                   User.email == usr['email']).first()
    if u is None:
        u = User(**usr)
        session.add(u)
        session.commit()
    return u


# Initialize the user
try:
    user = login()
except:
    user = None






# -*- coding: utf-8 -*-
import io
import os
import sys
import errno
from shutil import rmtree
from setuptools import setup, Command
from distutils.command.install import install

# Package meta-data.
NAME = 'pynorm'
DESCRIPTION = 'A probabilistic logic programming library for data science'
URL = 'https://github.com/reasoned-ai/norm'
EMAIL = 'xu@reasoned.ai'
AUTHOR = 'Xu Miao'
REQUIRES_PYTHON = '>=3.6.0'
VERSION = '0.1.1'

with io.open('requirements.txt') as f:
    REQUIRED = [line.split('==')[0] for line in f.readlines()]

EXTRAS = {}

with io.open('README.md', encoding='utf-8') as f:
    long_description = f.read()

BASEDIR = os.path.abspath(os.path.dirname(__file__))
PACKAGES = [
    'norm',
    'norm.db',
    'norm.executable',
    'norm.grammar',
    'norm.models'
]


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(BASEDIR, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPI via Twine…')
        os.system('twine upload dist/*')

        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(VERSION))
        os.system('git push --tags')

        sys.exit()


class InstallCommand(install):

    def run(self):
        install.run(self)
        home_dir = os.environ.get('NORM_HOME', '~/.norm')
        try:
            os.mkdir(home_dir)
            os.mkdir(os.path.join(home_dir, 'db'))
            os.mkdir(os.path.join(home_dir, 'data'))
            os.system('cp {}/norm/db/norm.db {}/db/norm.db'.format(BASEDIR, home_dir))
        except OSError as exc:
            if exc.errno != errno.EEXIST:
                raise
            print('\n\t\033[1;31;47m.norm\033[0m already exists, delete it first if you would like to overwrite\n')


setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=PACKAGES,
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    zip_safe=False,
    include_package_data=True,
    package_data={
        'norm': ['*.py', '*.db', '*.parq', '*.csv', '*.jsonl']
    },
    license='Apache 2.0',
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    # $ setup.py publish support.
    cmdclass={
        'upload': UploadCommand,
        'install': InstallCommand
    },
)



# Generated from /home/ax/Workspace/norm/norm/grammar/norm.g4 by ANTLR 4.7.2
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .normParser import normParser
else:
    from normParser import normParser

# This class defines a complete listener for a parse tree produced by normParser.
class normListener(ParseTreeListener):

    # Enter a parse tree produced by normParser#script.
    def enterScript(self, ctx:normParser.ScriptContext):
        pass

    # Exit a parse tree produced by normParser#script.
    def exitScript(self, ctx:normParser.ScriptContext):
        pass


    # Enter a parse tree produced by normParser#statement.
    def enterStatement(self, ctx:normParser.StatementContext):
        pass

    # Exit a parse tree produced by normParser#statement.
    def exitStatement(self, ctx:normParser.StatementContext):
        pass


    # Enter a parse tree produced by normParser#comments.
    def enterComments(self, ctx:normParser.CommentsContext):
        pass

    # Exit a parse tree produced by normParser#comments.
    def exitComments(self, ctx:normParser.CommentsContext):
        pass


    # Enter a parse tree produced by normParser#exports.
    def enterExports(self, ctx:normParser.ExportsContext):
        pass

    # Exit a parse tree produced by normParser#exports.
    def exitExports(self, ctx:normParser.ExportsContext):
        pass


    # Enter a parse tree produced by normParser#imports.
    def enterImports(self, ctx:normParser.ImportsContext):
        pass

    # Exit a parse tree produced by normParser#imports.
    def exitImports(self, ctx:normParser.ImportsContext):
        pass


    # Enter a parse tree produced by normParser#commands.
    def enterCommands(self, ctx:normParser.CommandsContext):
        pass

    # Exit a parse tree produced by normParser#commands.
    def exitCommands(self, ctx:normParser.CommandsContext):
        pass


    # Enter a parse tree produced by normParser#context.
    def enterContext(self, ctx:normParser.ContextContext):
        pass

    # Exit a parse tree produced by normParser#context.
    def exitContext(self, ctx:normParser.ContextContext):
        pass


    # Enter a parse tree produced by normParser#typeName.
    def enterTypeName(self, ctx:normParser.TypeNameContext):
        pass

    # Exit a parse tree produced by normParser#typeName.
    def exitTypeName(self, ctx:normParser.TypeNameContext):
        pass


    # Enter a parse tree produced by normParser#variable.
    def enterVariable(self, ctx:normParser.VariableContext):
        pass

    # Exit a parse tree produced by normParser#variable.
    def exitVariable(self, ctx:normParser.VariableContext):
        pass


    # Enter a parse tree produced by normParser#argumentProperty.
    def enterArgumentProperty(self, ctx:normParser.ArgumentPropertyContext):
        pass

    # Exit a parse tree produced by normParser#argumentProperty.
    def exitArgumentProperty(self, ctx:normParser.ArgumentPropertyContext):
        pass


    # Enter a parse tree produced by normParser#argumentDeclaration.
    def enterArgumentDeclaration(self, ctx:normParser.ArgumentDeclarationContext):
        pass

    # Exit a parse tree produced by normParser#argumentDeclaration.
    def exitArgumentDeclaration(self, ctx:normParser.ArgumentDeclarationContext):
        pass


    # Enter a parse tree produced by normParser#argumentDeclarations.
    def enterArgumentDeclarations(self, ctx:normParser.ArgumentDeclarationsContext):
        pass

    # Exit a parse tree produced by normParser#argumentDeclarations.
    def exitArgumentDeclarations(self, ctx:normParser.ArgumentDeclarationsContext):
        pass


    # Enter a parse tree produced by normParser#rename.
    def enterRename(self, ctx:normParser.RenameContext):
        pass

    # Exit a parse tree produced by normParser#rename.
    def exitRename(self, ctx:normParser.RenameContext):
        pass


    # Enter a parse tree produced by normParser#renames.
    def enterRenames(self, ctx:normParser.RenamesContext):
        pass

    # Exit a parse tree produced by normParser#renames.
    def exitRenames(self, ctx:normParser.RenamesContext):
        pass


    # Enter a parse tree produced by normParser#typeDeclaration.
    def enterTypeDeclaration(self, ctx:normParser.TypeDeclarationContext):
        pass

    # Exit a parse tree produced by normParser#typeDeclaration.
    def exitTypeDeclaration(self, ctx:normParser.TypeDeclarationContext):
        pass


    # Enter a parse tree produced by normParser#version.
    def enterVersion(self, ctx:normParser.VersionContext):
        pass

    # Exit a parse tree produced by normParser#version.
    def exitVersion(self, ctx:normParser.VersionContext):
        pass


    # Enter a parse tree produced by normParser#queryProjection.
    def enterQueryProjection(self, ctx:normParser.QueryProjectionContext):
        pass

    # Exit a parse tree produced by normParser#queryProjection.
    def exitQueryProjection(self, ctx:normParser.QueryProjectionContext):
        pass


    # Enter a parse tree produced by normParser#constant.
    def enterConstant(self, ctx:normParser.ConstantContext):
        pass

    # Exit a parse tree produced by normParser#constant.
    def exitConstant(self, ctx:normParser.ConstantContext):
        pass


    # Enter a parse tree produced by normParser#code.
    def enterCode(self, ctx:normParser.CodeContext):
        pass

    # Exit a parse tree produced by normParser#code.
    def exitCode(self, ctx:normParser.CodeContext):
        pass


    # Enter a parse tree produced by normParser#codeExpression.
    def enterCodeExpression(self, ctx:normParser.CodeExpressionContext):
        pass

    # Exit a parse tree produced by normParser#codeExpression.
    def exitCodeExpression(self, ctx:normParser.CodeExpressionContext):
        pass


    # Enter a parse tree produced by normParser#argumentExpression.
    def enterArgumentExpression(self, ctx:normParser.ArgumentExpressionContext):
        pass

    # Exit a parse tree produced by normParser#argumentExpression.
    def exitArgumentExpression(self, ctx:normParser.ArgumentExpressionContext):
        pass


    # Enter a parse tree produced by normParser#argumentExpressions.
    def enterArgumentExpressions(self, ctx:normParser.ArgumentExpressionsContext):
        pass

    # Exit a parse tree produced by normParser#argumentExpressions.
    def exitArgumentExpressions(self, ctx:normParser.ArgumentExpressionsContext):
        pass


    # Enter a parse tree produced by normParser#evaluationExpression.
    def enterEvaluationExpression(self, ctx:normParser.EvaluationExpressionContext):
        pass

    # Exit a parse tree produced by normParser#evaluationExpression.
    def exitEvaluationExpression(self, ctx:normParser.EvaluationExpressionContext):
        pass


    # Enter a parse tree produced by normParser#slicedExpression.
    def enterSlicedExpression(self, ctx:normParser.SlicedExpressionContext):
        pass

    # Exit a parse tree produced by normParser#slicedExpression.
    def exitSlicedExpression(self, ctx:normParser.SlicedExpressionContext):
        pass


    # Enter a parse tree produced by normParser#arithmeticExpression.
    def enterArithmeticExpression(self, ctx:normParser.ArithmeticExpressionContext):
        pass

    # Exit a parse tree produced by normParser#arithmeticExpression.
    def exitArithmeticExpression(self, ctx:normParser.ArithmeticExpressionContext):
        pass


    # Enter a parse tree produced by normParser#conditionExpression.
    def enterConditionExpression(self, ctx:normParser.ConditionExpressionContext):
        pass

    # Exit a parse tree produced by normParser#conditionExpression.
    def exitConditionExpression(self, ctx:normParser.ConditionExpressionContext):
        pass


    # Enter a parse tree produced by normParser#oneLineExpression.
    def enterOneLineExpression(self, ctx:normParser.OneLineExpressionContext):
        pass

    # Exit a parse tree produced by normParser#oneLineExpression.
    def exitOneLineExpression(self, ctx:normParser.OneLineExpressionContext):
        pass


    # Enter a parse tree produced by normParser#multiLineExpression.
    def enterMultiLineExpression(self, ctx:normParser.MultiLineExpressionContext):
        pass

    # Exit a parse tree produced by normParser#multiLineExpression.
    def exitMultiLineExpression(self, ctx:normParser.MultiLineExpressionContext):
        pass


    # Enter a parse tree produced by normParser#none.
    def enterNone(self, ctx:normParser.NoneContext):
        pass

    # Exit a parse tree produced by normParser#none.
    def exitNone(self, ctx:normParser.NoneContext):
        pass


    # Enter a parse tree produced by normParser#bool_c.
    def enterBool_c(self, ctx:normParser.Bool_cContext):
        pass

    # Exit a parse tree produced by normParser#bool_c.
    def exitBool_c(self, ctx:normParser.Bool_cContext):
        pass


    # Enter a parse tree produced by normParser#integer_c.
    def enterInteger_c(self, ctx:normParser.Integer_cContext):
        pass

    # Exit a parse tree produced by normParser#integer_c.
    def exitInteger_c(self, ctx:normParser.Integer_cContext):
        pass


    # Enter a parse tree produced by normParser#float_c.
    def enterFloat_c(self, ctx:normParser.Float_cContext):
        pass

    # Exit a parse tree produced by normParser#float_c.
    def exitFloat_c(self, ctx:normParser.Float_cContext):
        pass


    # Enter a parse tree produced by normParser#string_c.
    def enterString_c(self, ctx:normParser.String_cContext):
        pass

    # Exit a parse tree produced by normParser#string_c.
    def exitString_c(self, ctx:normParser.String_cContext):
        pass


    # Enter a parse tree produced by normParser#pattern.
    def enterPattern(self, ctx:normParser.PatternContext):
        pass

    # Exit a parse tree produced by normParser#pattern.
    def exitPattern(self, ctx:normParser.PatternContext):
        pass


    # Enter a parse tree produced by normParser#uuid.
    def enterUuid(self, ctx:normParser.UuidContext):
        pass

    # Exit a parse tree produced by normParser#uuid.
    def exitUuid(self, ctx:normParser.UuidContext):
        pass


    # Enter a parse tree produced by normParser#url.
    def enterUrl(self, ctx:normParser.UrlContext):
        pass

    # Exit a parse tree produced by normParser#url.
    def exitUrl(self, ctx:normParser.UrlContext):
        pass


    # Enter a parse tree produced by normParser#datetime.
    def enterDatetime(self, ctx:normParser.DatetimeContext):
        pass

    # Exit a parse tree produced by normParser#datetime.
    def exitDatetime(self, ctx:normParser.DatetimeContext):
        pass


    # Enter a parse tree produced by normParser#logicalOperator.
    def enterLogicalOperator(self, ctx:normParser.LogicalOperatorContext):
        pass

    # Exit a parse tree produced by normParser#logicalOperator.
    def exitLogicalOperator(self, ctx:normParser.LogicalOperatorContext):
        pass


    # Enter a parse tree produced by normParser#spacedLogicalOperator.
    def enterSpacedLogicalOperator(self, ctx:normParser.SpacedLogicalOperatorContext):
        pass

    # Exit a parse tree produced by normParser#spacedLogicalOperator.
    def exitSpacedLogicalOperator(self, ctx:normParser.SpacedLogicalOperatorContext):
        pass


    # Enter a parse tree produced by normParser#newlineLogicalOperator.
    def enterNewlineLogicalOperator(self, ctx:normParser.NewlineLogicalOperatorContext):
        pass

    # Exit a parse tree produced by normParser#newlineLogicalOperator.
    def exitNewlineLogicalOperator(self, ctx:normParser.NewlineLogicalOperatorContext):
        pass


    # Enter a parse tree produced by normParser#conditionOperator.
    def enterConditionOperator(self, ctx:normParser.ConditionOperatorContext):
        pass

    # Exit a parse tree produced by normParser#conditionOperator.
    def exitConditionOperator(self, ctx:normParser.ConditionOperatorContext):
        pass


    # Enter a parse tree produced by normParser#spacedConditionOperator.
    def enterSpacedConditionOperator(self, ctx:normParser.SpacedConditionOperatorContext):
        pass

    # Exit a parse tree produced by normParser#spacedConditionOperator.
    def exitSpacedConditionOperator(self, ctx:normParser.SpacedConditionOperatorContext):
        pass





"""A collection of ORM sqlalchemy models for CoreLambda"""
import json

import pandas as pd
from pandas import DataFrame

from norm.models import Register
from norm.models.norm import Lambda, Variable, Status

import logging
logger = logging.getLogger(__name__)


class CoreLambda(Lambda):
    """
    Core functions are at the computable level and predefined in the code base
    """
    __mapper_args__ = {
        'polymorphic_identity': 'lambda_core'
    }
    NAMESPACE = 'norm.core'

    def __init__(self, name, description, variables, dtype='object'):
        super().__init__(namespace=self.NAMESPACE,
                         name=name,
                         description=description,
                         variables=variables,
                         dtype=dtype)
        self.status = Status.READY
        self.atomic = True


@Register()
class ReadFileLambda(CoreLambda):
    __mapper_args__ = {
        'polymorphic_identity': 'lambda_core_read_file'
    }

    VAR_PATH = 'path'
    VAR_PARM = 'parameters'
    VAR_EXT = 'ext'

    EXT_CSV = 'csv'
    EXT_PAR = 'parq'
    EXT_TSV = 'tsv'
    EXT_JSL = 'jsonl'

    def __init__(self):
        from norm.models import lambdas
        super().__init__(name='read',
                         description='Read data from files [.csv, .tsv, .parq, .jsonl], default to csv file'
                                     'your_lambda.read("path_to_the_file.csv", (sep="\t", skiprows=3))'
                                     'read(your_lambda, "path_to_the_file.csv", (sep="\t", skiprows=3))',
                         variables=[Variable(self.VAR_PATH, lambdas.String),
                                    Variable(self.VAR_PARM, lambdas.Any),
                                    Variable(self.VAR_EXT, lambdas.String),
                                    Variable(self.VAR_OUTPUT, lambdas.Any)])

    def __call__(self, **inputs):
        path = inputs.get(self.VAR_PATH)
        params = inputs.get(self.VAR_PARM)
        ext = inputs.get(self.VAR_EXT)
        if path is None:
            return None

        if isinstance(path, DataFrame):
            path = path.iloc[0][0]
        assert(isinstance(path, str))
        if ext and isinstance(ext, DataFrame):
            ext = ext.iloc[0][0]
        if ext is None:
            last_path = path.split('.')[-1]
            if last_path == self.EXT_CSV:
                ext = self.EXT_CSV
            elif last_path == self.EXT_PAR:
                ext = self.EXT_PAR
            elif last_path == self.EXT_TSV:
                ext = self.EXT_TSV
            elif last_path == self.EXT_JSL:
                ext = self.EXT_JSL
            else:
                ext = self.EXT_CSV
        if params is not None and isinstance(params, DataFrame):
            params = params.to_dict()[0]
        else:
            params = {}

        if ext == self.EXT_CSV:
            return self.read_csv(path, params)
        elif ext == self.EXT_TSV:
            params['sep'] = '\t'
            return self.read_csv(path, params)
        elif ext == self.EXT_PAR:
            return self.read_parquet(path, params)
        elif ext == self.EXT_JSL:
            return self.read_jsonl(path)
        else:
            msg = 'Currently supported file formats: CSV (.csv), TSV (.tsv), Parquet (.parq) ' \
                  'and JSONL (.jsonl)'
            logger.error(msg)
            raise TypeError(msg)

    def read_csv(self, path, params):
        return pd.read_csv(path, **params)

    def read_parquet(self, path, params):
        return pd.read_parquet(path, **params)

    def read_jsonl(self, path):
        with open(path) as f:
            return DataFrame([json.loads(line) for line in f])


@Register()
class StringFormatterLambda(CoreLambda):
    __mapper_args__ = {
        'polymorphic_identity': 'lambda_core_string_formatter'
    }

    VAR_FORMATTER = 'formatter'
    VAR_VARIABLES = 'variables'

    def __init__(self):
        from norm.models import lambdas
        formatter = Variable(self.VAR_FORMATTER, lambdas.String)
        variables = Variable(self.VAR_VARIABLES, lambdas.Any)
        output = Variable(self.VAR_OUTPUT, lambdas.String)
        super().__init__(name='format',
                         description='Format the strings with given inputs, the semantic is the same as Python format',
                         variables=[formatter, variables, output])

    def __call__(self, **inputs):
        formatter = inputs.get(self.VAR_FORMATTER, None)
        variables = inputs.get(self.VAR_VARIABLES, None)
        if variables is None or formatter is None:
            return None

        v_cols = list(variables.columns)
        if isinstance(formatter, DataFrame):
            assert(len(formatter) == len(variables))
            f_col = formatter.columns[0]
            df = formatter.reset_index().join(variables.reset_index())
            data = DataFrame(df.apply(lambda x: x[f_col].format(*[x[c] for c in v_cols]), axis=1),
                             columns=[self.VAR_OUTPUT])
        else:
            assert(isinstance(formatter, str))
            data = DataFrame(variables.apply(lambda x: formatter.format(*[x[c] for c in v_cols]), axis=1),
                             columns=[self.VAR_OUTPUT])
        return data


@Register()
class ExtractPatternLambda(CoreLambda):
    __mapper_args__ = {
        'polymorphic_identity': 'lambda_core_extract_pattern'
    }

    VAR_PATTERN = 'pattern'
    VAR_STRING = 'string'
    VAR_FILLNA = 'fillna'

    def __init__(self):
        from norm.models import lambdas
        super().__init__(name='extract',
                         description='Extract patterns from a string'
                                     '"(2014).*(6)".extract("2014-06") --> (2014, 6)/None'
                                     '"2014.*6".extract(s, fillna=False) --> True/False',
                         variables=[Variable(self.VAR_PATTERN, lambdas.String),
                                    Variable(self.VAR_STRING, lambdas.String),
                                    Variable(self.VAR_FILLNA, lambdas.Any),
                                    Variable(self.VAR_OUTPUT, lambdas.Any)])

    def __call__(self, **inputs):
        pattern = inputs.get(self.VAR_PATTERN, None)
        string = inputs.get(self.VAR_STRING, None)
        if pattern is None or string is None:
            return None
        fill_na = inputs.get(self.VAR_FILLNA, None)
        if fill_na is not None:
            fill_na = fill_na.data

        import re

        def match(p, s):
            m = re.match(p, s)
            if m is not None:
                if len(m.groups()) == 0:
                    return True
                else:
                    return m.groups()
            else:
                return fill_na

        if isinstance(pattern, DataFrame):
            assert(len(pattern) == len(string))
            pdata = pattern.loc[:, 0]
            sdata = string.loc[:, 0]
            df = DataFrame({'p': pdata, 's': sdata})
            data = df.apply(lambda x: match(x.p, x.s), axis=1, result_type='expand').drop(columns=['p', 's'])
        else:
            assert(isinstance(pattern, str))
            f = pattern
            data = string.apply(lambda x: match(f, x.s), axis=1, result_type='expand')
        return data



class NormError(RuntimeError):
    pass


class NormExecutable(object):

    VARIABLE_SEPARATOR = '__dot__'

    def __init__(self):
        from norm.models.norm import Lambda
        self.lam: Lambda = None

    def try_retrieve_type(self, session, namespaces, name, version=None, status=None):
        current_session_id = session.hash_key
        from norm.config import cache
        from norm.models.norm import retrieve_type, Lambda
        if isinstance(namespaces, str):
            key = namespaces
        else:
            key = ''.join(sorted(namespaces))
        key = (key, name, version, status)
        lam = cache.get(key)
        if lam is not None:
            if lam._sa_instance_state.session_id != current_session_id:
                new_lam: Lambda = session.merge(lam)
                # TODO: might have other states need to be kept
                new_lam._data = lam._data
                lam = new_lam
            return lam
        lam = retrieve_type(namespaces, name, version, status)
        cache[key] = lam
        return lam

    def compile(self, context):
        """
        Compile the command with the given context
        :param context: the context of the executable
        :type context: norm.engine.NormCompiler
        :return: An NormExecutable
        :rtype: NormExecutable
        """
        return self

    def execute(self, context):
        """
        Execute the command with given context
        :param context: the context of the executable
        :type context: norm.engine.NormCompiler
        :return: the result
        """
        return self.lam


class Projection(NormExecutable):

    def __init__(self, variables, to_evaluate=False):
        """
        The projection definition
        :param variables: a list of variables to project on
        :type variables: List[norm.executable.variable.VariableName]
        :param to_evaluate: whether to evaluate these variables or not, default to False.
        :type to_evaluate: Boolean
        """
        super().__init__()
        self.variables = variables
        self.to_evaluate = to_evaluate

    @property
    def num(self):
        return len(self.variables)






from pandas import DataFrame
from hashids import Hashids
from zlib import adler32
h = Hashids()


def hash_df(df):
    """
    Create a hash string out of a DataFrame data
    :param df: the DataFrame data
    :type df: DataFrame
    :return: the hash string
    :rtype: str
    """
    if df is None:
        return ''
    # noinspection PyUnresolvedReferences
    return h.encode(adler32(str(df.values).encode('utf-8')))



import enum
from norm.executable import NormExecutable
from norm.executable import Projection
from pandas import DataFrame, Series, concat, Index


class Strategy(enum.Enum):
    """
    Strategy to unify inputs:
        shortest: align the data and cut to the shortest length. If two inputs, one with a list of 10 objects, the
                  other with a list of 2 objects. The unified data contains 2 objects.
        longest: align the data to the longest length. Fill with n/a or type defaults.
    """
    SHORTEST = 'shortest'
    LONGEST = 'longest'


class NormExpression(NormExecutable):

    def __init__(self):
        super().__init__()
        self.projection: Projection = None
        self.eval_lam = None
        self.description: str = None

    def unify(self, inputs, strategy=Strategy.LONGEST):
        """
        Unify multi-variates into one DataFrame.
        :param inputs: the dictionary of the input values
        :type inputs: Dict
        :param strategy: the strategy to align inputs
        :type strategy: Strategy
        :return: the unified DataFrame
        :rtype: DataFrame
        """
        from norm.models.norm import Lambda
        lambda_inputs = {}
        list_inputs = {}
        dataframe_inputs = {}
        constant_inputs = {}
        for k, v in inputs.items():
            if isinstance(v, Lambda):
                lambda_inputs[k] = v
            elif isinstance(v, list):
                list_inputs[k] = v
            elif isinstance(v, (Series, Index)):
                list_inputs[k] = v.values
            elif isinstance(v, DataFrame):
                dataframe_inputs[k] = v
            else:
                constant_inputs[k] = v

        # fill list inputs with None
        if len(list_inputs) > 0:
            max_len = max(len(v) for k, v in list_inputs.items())
            for k, v in list_inputs.items():
                if len(v) < max_len:
                    if not isinstance(v, list):
                        v = list(v)
                        list_inputs[k] = v
                    v.extend([None] * (max_len - len(v)))

        # concat dataframe inputs
        to_concat = [DataFrame(list_inputs)]
        for k, v in dataframe_inputs.items():
            if len(v.columns) == 1:
                renames = {col: k for col in v.columns}
            else:
                renames = {col: '{}{}{}'.format(k, self.VARIABLE_SEPARATOR, col) for col in v.columns}
            to_concat.append(v.rename(columns=renames))

        # concat constant inputs
        rt = concat(to_concat, axis=1)
        if len(rt) == 0:
            for k, v in constant_inputs.items():
                rt.loc[0, k] = v
        else:
            for k, v in constant_inputs.items():
                rt[k] = v
        return rt

    def execute(self, context):
        return self.lam.data



"""A collection of ORM sqlalchemy models for NativeLambda"""
from norm.models import Register
from norm.models.norm import Lambda, Variable, Status

from datetime import datetime
import numpy as np

import logging
logger = logging.getLogger(__name__)


class NativeLambda(Lambda):
    __mapper_args__ = {
        'polymorphic_identity': 'lambda_native'
    }
    NAMESPACE = 'norm.native'

    def __init__(self, name, description, variables, dtype='object'):
        super().__init__(namespace=self.NAMESPACE,
                         name=name,
                         description=description,
                         variables=variables,
                         dtype=dtype)
        self.status = Status.READY
        self.atomic = True

    def empty_data(self):
        return None


@Register()
class TypeLambda(NativeLambda):
    __mapper_args__ = {
        'polymorphic_identity': 'lambda_native_type'
    }

    def __init__(self):
        super().__init__(name='Type',
                         description='A logical function',
                         variables=[])


@Register()
class AnyLambda(NativeLambda):
    __mapper_args__ = {
        'polymorphic_identity': 'lambda_native_any'
    }

    def __init__(self):
        super().__init__(name='Any',
                         description='Any type',
                         variables=[])

    @property
    def default(self):
        return None


class ListLambda(NativeLambda):
    __mapper_args__ = {
        'polymorphic_identity': 'lambda_native_list'
    }
    INTERN = 'intern'

    def __init__(self, type_):
        """
        :param type_: the intern type of the list
        :type type_: Lambda
        """
        variable = Variable(self.INTERN, type_)
        super().__init__(name='List[{}]'.format(type_.signature),
                         description='A list of a certain type',
                         variables=[variable])

    @property
    def default(self):
        return []


@Register()
class BooleanLambda(NativeLambda):
    __mapper_args__ = {
        'polymorphic_identity': 'lambda_native_boolean'
    }

    def __init__(self):
        super().__init__(name='Boolean',
                         description='Boolean, true/false',
                         variables=[],
                         dtype='bool')

    @property
    def default(self):
        return False


@Register()
class FloatLambda(NativeLambda):
    __mapper_args__ = {
        'polymorphic_identity': 'lambda_native_float'
    }

    def __init__(self):
        super().__init__(name='Float',
                         description='Integer, -inf..+inf',
                         variables=[],
                         dtype='float')

    @property
    def default(self):
        return 0.0


@Register()
class IntegerLambda(NativeLambda):
    __mapper_args__ = {
        'polymorphic_identity': 'lambda_native_integer'
    }

    def __init__(self):
        super().__init__(name='Integer',
                         description='Integer, -inf..+inf',
                         variables=[],
                         dtype='int')

    @property
    def default(self):
        return 0


@Register()
class StringLambda(NativeLambda):
    __mapper_args__ = {
        'polymorphic_identity': 'lambda_native_string'
    }

    def __init__(self):
        super().__init__(name='String',
                         description='String, "blahbalh"',
                         variables=[])

    @property
    def default(self):
        return ''


@Register()
class PatternLambda(NativeLambda):
    __mapper_args__ = {
        'polymorphic_identity': 'lambda_native_pattern'
    }

    def __init__(self):
        super().__init__(name='Pattern',
                         description='Pattern, r"^test[0-9]+"',
                         variables=[])

    @property
    def default(self):
        return ''


@Register()
class UUIDLambda(NativeLambda):
    __mapper_args__ = {
        'polymorphic_identity': 'lambda_native_uuid'
    }

    def __init__(self):
        super().__init__(name='UUID',
                         description='UUID, $"sfsfsfsf"',
                         variables=[])

    @property
    def default(self):
        return ''


@Register()
class URLLambda(NativeLambda):
    __mapper_args__ = {
        'polymorphic_identity': 'lambda_native_url'
    }

    def __init__(self):
        super().__init__(name='URL',
                         description='URL, l"http://example.com"',
                         variables=[])

    @property
    def default(self):
        return 'http://'


@Register()
class DatetimeLambda(NativeLambda):
    __mapper_args__ = {
        'polymorphic_identity': 'lambda_native_datetime'
    }

    def __init__(self):
        super().__init__(name='Datetime',
                         description='Datetime, t"2018-09-01"',
                         variables=[],
                         dtype='datetime64[ns]')

    @property
    def default(self):
        return np.datetime64(datetime.utcnow())


def get_type_by_dtype(dtype):
    """
    Convert the numpy dtype to native lambda
    :param dtype: the numpy dtype
    :type dtype: numpy.dtype
    :return: the Lambda
    :rtype: Lambda
    """
    from norm.models import lambdas
    if dtype.name.find('int') > -1:
        return lambdas.Integer
    elif dtype.name.find('float') > -1:
        return lambdas.Float
    else:
        return lambdas.Any
